unit KOLJpeg;

interface

uses Windows, KOL;

const
  JPEG_LIB_VERSION = 62;

{$A4}

type
  TJPEGScale = ( jsFullSize, jsHalf, jsQuarter, jsEighth );
  TJPEGPerformance = (jpBestQuality, jpBestSpeed);
  TJPEGQualityRange = 1..100;   // 100 = best quality, 25 = pretty awful
  TJPEGPixelFormat = (jf24Bit, jf8Bit);

  PJPEGData = ^TJPEGData;
  TJPEGData = object( TObj )
  private
    FData: PStream;
    FHeight: Integer;
    FWidth: Integer;
    FGrayscale: Boolean;
  protected
  public
    destructor Destroy; virtual;
    procedure Clear;
  end;

  PKOLJpeg = ^TKOLJpeg;

  TJpegProgress = procedure( JPEG_obj: PKOLJpeg; const R: TRect; var Stop: Boolean ) of object;
  TOnGetScanline = procedure( JPEG_obj: PKOLJpeg; var w, h, c: Integer;
    line: Integer; var Scanline: Pointer ) of object;

  TKOLJpeg = object( TObj )
  private
    FData: PStream;
    FImage: PJPEGData;
    FCorrupted: Boolean;
    FWidth, FHeight: Integer;
    FGrayScale: Boolean;
    FCMYK: Boolean;
    FProgressiveEncoding: Boolean;
    FBitmap: PBitmap;
    FCallback: Pointer;
    FScale: TJPEGScale;
    FSmoothing: Boolean;
    FPerformance: TJPEGPerformance;
    FProgressiveDisplay: Boolean;
    Jpeg_Error: Boolean;
    FStop: Boolean;
    FProgress: function( JPEG_obj: PKOLJpeg; const R: TRect ): Boolean;
    FTempPal: HPalette;
    FConvertCMYK2RGBProc: procedure( Bmp: PBitmap; Y_start, Y_end: Integer );
    FConvertCMYK2RGB: Boolean;
    // for compression:
    FQuality: TJPEGQualityRange;
    FPixelFormat: TJPEGPixelFormat;
    fProgressTime: Integer;
    FOnProgress: TJPEGProgress;
    FOnGetScanline: TOnGetScanline;
    procedure CreateImage;
    function GetBitmap: PBitmap;
    procedure SetBitmap(const Value: PBitmap);
    procedure Changed;
    procedure SetConvertCMYK2RGB(const Value: Boolean);
    procedure SetOnProgress(const Value: TJPEGProgress);
    procedure SetScale(const Value: TJPEGScale);
  protected
    procedure Compress;
  protected
    procedure Init; virtual;
  public
    destructor Destroy; virtual;
    function LoadFromStream( Strm: PStream ): Boolean;
    function LoadFromFile( const filename: String ): Boolean;
    property Corrupted: Boolean read FCorrupted;
    procedure Clear;
    property GrayScale: Boolean read FGrayScale write FGrayScale;
    property ProgressiveEncoding: Boolean read FProgressiveEncoding write FProgressiveEncoding;
    property Bitmap: PBitmap read GetBitmap write SetBitmap;
    procedure DIBNeeded;
    property Width: Integer read FWidth;
    property Height: Integer read FHeight;
    property Scale: TJPEGScale read FScale write SetScale;
    property Smoothing: Boolean read FSmoothing write FSmoothing;
    property Performance: TJPEGPerformance read FPerformance write FPerformance;
    property ProgressiveDisplay: Boolean read FProgressiveDisplay write FProgressiveDisplay;
    property CMYK: Boolean read FCMYK;
    property ConvertCMYK2RGB: Boolean read FConvertCMYK2RGB write SetConvertCMYK2RGB;
    property PixelFormat: TJPEGPixelFormat read FPixelFormat write FPixelFormat;
    function Empty: Boolean;
    function HasBitmap: Boolean;
    procedure FreeBitmap;

    procedure Draw( DC: HDC; x, y: Integer );
    procedure StretchDraw( DC: HDC; const R: TRect );

    property CompressionQuality: TJPEGQualityRange read FQuality write FQuality;
    procedure JPEGNeeded;
    property OnGetScanline: TOnGetScanline read FOnGetScanline write FOnGetScanline;

    function SaveToFile( const filename: String ): Boolean;
    function SaveToStream( Strm: PStream ): Boolean;
    property OnProgress: TJPEGProgress read FOnProgress write SetOnProgress;
  end;

function NewJpeg: PKOLJpeg;

type
  TJPEGDefaults = record
    CompressionQuality: TJPEGQualityRange;
    Grayscale: Boolean;
    Performance: TJPEGPerformance;
    PixelFormat: TJPEGPixelFormat; // only for compression
    ProgressiveDisplay: Boolean;
    ProgressiveEncoding: Boolean;
    Scale: TJPEGScale;
    Smoothing: Boolean;
  end;

var   // Default settings for all new TJPEGImage instances
  JPEGDefaults: TJPEGDefaults = (
    CompressionQuality: 90;
    Grayscale: False;
    Performance: jpBestQuality;
    PixelFormat: jf24Bit;         // initialized to match video mode
    ProgressiveDisplay: False;
    ProgressiveEncoding: False;
    Scale: jsFullSize;
    Smoothing: True;
  );

procedure DoConvertCMYK2RGB( Bmp: PBitmap; Y_start, Y_end: Integer );

implementation

uses jerror, jdapimin, jdatasrc, jcomapi, jdapistd,
     jcapimin, jdatadst, jcparam, jcapistd;

const
  DCTSIZE2 = 64;
  NUM_QUANT_TBLS = 4;
  NUM_HUFF_TBLS  = 4;
  NUM_ARITH_TBLS = 16;
  MAX_COMPS_IN_SCAN = 4;
  D_MAX_BLOCKS_IN_MCU = 10;
  C_MAX_BLOCKS_IN_MCU = 10;     { compressor's limit on blocks per MCU }

type
//* Known color spaces. */
  J_COLOR_SPACE = (
	JCS_UNKNOWN,		//* error/unspecified */
	JCS_GRAYSCALE,		//* monochrome */
	JCS_RGB,		//* red/green/blue */
	JCS_YCbCr,		//* Y/Cb/Cr (also known as YUV) */
	JCS_CMYK,		//* C/M/Y/K */
	JCS_YCCK		//* Y/Cb/Cr/K */
  );

//* DCT/IDCT algorithm options. */
  J_DCT_METHOD = (
	JDCT_ISLOW,		//* slow but accurate integer algorithm */
	JDCT_IFAST,		//* faster, less accurate integer method */
	JDCT_FLOAT		//* floating-point: accurate, fast on fast HW */
  );

//* Dithering options for decompression. */
  J_DITHER_MODE = (
	JDITHER_NONE,		//* no dithering */
	JDITHER_ORDERED,	//* simple ordered dither */
	JDITHER_FS		//* Floyd-Steinberg error diffusion dither */
  );

  JSAMPLE = BYTE;
  JSAMPLEARRAY = array[0..255] of JSAMPLE;
  PJSAMPROW = ^JSAMPLEArray;
  JSAMPARRAY = array[0..255] of PJSAMPROW;
  PJSAMPARRAY = ^JSAMPARRAY;

  JCOEFF = SmallInt;
  JBLOCK = array[0..DCTSIZE2-1] of JCOEFF;
  JBLOCKROW = ^JBLOCK;
  JBLOCKARRAY = ^JBLOCKROW;

//* DCT coefficient quantization tables. */
  JQUANT_TBL = record
    (* This array gives the coefficient quantizers in natural array order
     * (not the zigzag order in which they are stored in a JPEG DQT marker).
     * CAUTION: IJG versions prior to v6a kept this array in zigzag order.
     *)
    quantval: array[0..DCTSIZE2-1] of WORD; //* quantization step for each coefficient */
    (* This field is used only during compression.  It's initialized FALSE when
     * the table is created, and set TRUE when it's been output to the file.
     * You could suppress output of a table by setting this to TRUE.
     * (See jpeg_suppress_tables for an example.)
     *)
    sent_table: Boolean;		//* TRUE when table has been output */
  end;
  PJQUANT_TBL = ^JQUANT_TBL;

//* Huffman coding tables. */
  JHUFF_TBL = record
    //* These two fields directly represent the contents of a JPEG DHT marker */
    bits: array[0..16] of BYTE;	//* bits[k] = # of symbols with codes of */
                                  //* length k bits; bits[0] is unused */
    huffval: array[0..255] of BYTE;//* The symbols, in order of incr code length */
    (* This field is used only during compression.  It's initialized FALSE when
     * the table is created, and set TRUE when it's been output to the file.
     * You could suppress output of a table by setting this to TRUE.
     * (See jpeg_suppress_tables for an example.)
     *)
    sent_table: Boolean;		//* TRUE when table has been output */
  end;
  PJHUFF_TBL = ^JHUFF_TBL;

//* Basic info about one component (color channel). */
  jpeg_component_info = record
    //* These values are fixed over the whole image. */
    //* For compression, they must be supplied by parameter setup; */
    //* for decompression, they are read from the SOF marker. */
    component_id: Integer;	//* identifier for this component (0..255) */
    component_index: Integer;	//* its index in SOF or cinfo->comp_info[] */
    h_samp_factor: Integer;	//* horizontal sampling factor (1..4) */
    v_samp_factor: Integer;	//* vertical sampling factor (1..4) */
    quant_tbl_no: Integer;	//* quantization table selector (0..3) */
    //* These values may vary between scans. */
    //* For compression, they must be supplied by parameter setup; */
    //* for decompression, they are read from the SOS marker. */
    //* The decompressor output side may not use these variables. */
    dc_tbl_no: Integer;		//* DC entropy table selector (0..3) */
    ac_tbl_no: Integer;		//* AC entropy table selector (0..3) */

    //* Remaining fields should be treated as private by applications. */

    //* These values are computed during compression or decompression startup: */
    (* Component's size in DCT blocks.
     * Any dummy blocks added to complete an MCU are not counted; therefore
     * these values do not depend on whether a scan is interleaved or not.
     *)
    width_in_blocks: DWORD;
    height_in_blocks: DWORD;
    (* Size of a DCT block in samples.  Always DCTSIZE for compression.
     * For decompression this is the size of the output from one DCT block,
     * reflecting any scaling we choose to apply during the IDCT step.
     * Values of 1,2,4,8 are likely to be supported.  Note that different
     * components may receive different IDCT scalings.
     *)
    DCT_scaled_size: Integer;
    (* The downsampled dimensions are the component's actual, unpadded number
     * of samples at the main buffer (preprocessing/compression interface), thus
     * downsampled_width = ceil(image_width * Hi/Hmax)
     * and similarly for height.  For decompression, IDCT scaling is included, so
     * downsampled_width = ceil(image_width * Hi/Hmax * DCT_scaled_size/DCTSIZE)
     *)
    downsampled_width: DWORD;	 //* actual width in samples */
    downsampled_height: DWORD;     //* actual height in samples */
    (* This flag is used only for decompression.  In cases where some of the
     * components will be ignored (eg grayscale output from YCbCr image),
     * we can skip most computations for the unused components.
     *)
    component_needed: Boolean;	//* do we need the value of this component? */

    //* These values are computed before starting a scan of the component. */
    //* The decompressor output side may not use these variables. */
    MCU_width: Integer;		//* number of blocks per MCU, horizontally */
    MCU_height: Integer;		//* number of blocks per MCU, vertically */
    MCU_blocks: Integer;		//* MCU_width * MCU_height */
    MCU_sample_width: Integer;	//* MCU width in samples, MCU_width*DCT_scaled_size */
    last_col_width: Integer;	//* # of non-dummy blocks across in last MCU */
    last_row_height: Integer;	//* # of non-dummy blocks down in last MCU */

    (* Saved quantization table for component; NULL if none yet saved.
     * See jdinput.c comments about the need for this information.
     * This field is currently used only for decompression.
     *)
    quant_table: PJQUANT_TBL;

    //* Private per-component storage for DCT or IDCT subsystem. */
    dct_table: Pointer;
  end;

  Pjpeg_component_info = ^jpeg_component_info;

//* Error handler object */

  jpeg_error_mgr = record
    //* Error exit handler: does not return to caller */
    error_exit: procedure( cinfo: Pointer ); cdecl;
    //* Conditionally emit a trace or warning message */
    emit_message: procedure( cinfo: Pointer; msg_level: Integer ); cdecl;
    //* Routine that actually outputs a trace or error message */
    output_message: procedure( cinfo: Pointer ); cdecl;
    //* Format a message string for the most recent JPEG error or message */
    format_message: procedure( cinfo: Pointer; buffer: PChar ); cdecl;
  //#define JMSG_LENGTH_MAX  200	/* recommended size of format_message buffer */
    //* Reset error state variables at start of a new image */
    reset_error_mgr: procedure( cinfo: Pointer ); cdecl;

    (* The message ID code and any parameters are saved here.
     * A message can have one string parameter or up to 8 int parameters.
     *)
    msg_code: Integer;
  //#define JMSG_STR_PARM_MAX  80
    (*union {
      int i[8];
      char s[JMSG_STR_PARM_MAX];
    } msg_parm;*)
    msg_param: array[ 0..79 ] of Char;

    //* Standard state variables for error facility */

    trace_level: Integer;		//* max msg_level that will be displayed */

    (* For recoverable corrupt-data errors, we emit a warning message,
     * but keep going unless emit_message chooses to abort.  emit_message
     * should count warnings in num_warnings.  The surrounding application
     * can check for bad data by seeing if num_warnings is nonzero at the
     * end of processing.
     *)
    num_warnings: Integer;	//* number of corrupt-data warnings */

    (* These fields point to the table(s) of error message strings.
     * An application can change the table pointer to switch to a different
     * message list (typically, to change the language in which errors are
     * reported).  Some applications may wish to add additional error codes
     * that will be handled by the JPEG library error mechanism; the second
     * table pointer is used for this purpose.
     *
     * First table includes all errors generated by JPEG library itself.
     * Error code 0 is reserved for a "no such error string" message.
     *)
    jpeg_message_table: PPChar;   //* Library errors */
    last_jpeg_message: Integer;   //* Table contains strings 0..last_jpeg_message */
    (* Second table can be added by application (see cjpeg/djpeg for example).
     * It contains strings numbered first_addon_message..last_addon_message.
     *)
    addon_message_table: PPchar;  //* Non-library errors */
    first_addon_message: Integer;	//* code for first string in addon table */
    last_addon_message: Integer;	//* code for last string in addon table */
  end;

  Pjpeg_error_mgr = ^jpeg_error_mgr;

(* Memory manager object.
 * Allocates "small" objects (a few K total), "large" objects (tens of K),
 * and "really big" objects (virtual arrays with backing store if needed).
 * The memory manager does not allow individual objects to be freed; rather,
 * each created object is assigned to a pool, and whole pools can be freed
 * at once.  This is faster and more convenient than remembering exactly what
 * to free, especially where malloc()/free() are not too speedy.
 * NB: alloc routines never return NULL.  They exit to error_exit if not
 * successful.
 *)

//#define JPOOL_PERMANENT	0	/* lasts until master record is destroyed */
//#define JPOOL_IMAGE	1	/* lasts until done with image/datastream */
//#define JPOOL_NUMPOOLS	2

//typedef struct jvirt_sarray_control * jvirt_sarray_ptr;
//typedef struct jvirt_barray_control * jvirt_barray_ptr;

  jpeg_memory_mgr = record
    //* Method pointers */
    alloc_small: function( cinfo: Pointer; pool_id: Integer;
                 sizeofobject: Integer ): Pointer; cdecl;
    alloc_large: function( cinfo: Pointer; pool_id: Integer;
                 sizeofobject: Integer ): Pointer; cdecl;
    alloc_sarray: function( cinfo: Pointer; pool_id: Integer; samplesperrow,
                  numrows: DWORD ): PJSAMPARRAY; cdecl;
    alloc_barray: function( cinfo: Pointer; pool_id: Integer; blocksperrow,
                  numrows: DWORD ): JBLOCKARRAY; cdecl;
    request_virt_sarray: function( cinfo: Pointer; pool_id: Integer;
                         prezero: Boolean; samplesperrow, numrows,
                         maxaccess: DWORD ): Pointer; cdecl;
    request_virt_barray: function( cinfo: Pointer; pool_id: Integer;
                         prezero: Boolean; blocksperrow, numrows,
                         maxaccess: DWORD ): Pointer; cdecl;
    realize_virt_arrays: procedure( cinfo: Pointer ); cdecl;
    access_virt_sarray: function( cinfo: Pointer; ptr: Pointer;
                        start_row, num_rows: DWORD;
                        writable: Boolean ): PJSAMPARRAY; cdecl;
    access_virt_barray: function( cinfo: Pointer; ptr: Pointer;
                        start_row, num_rows: DWORD;
                        writable: Boolean ): JBLOCKARRAY; cdecl;
    free_pool: procedure( cinfo: Pointer; pool_id: Integer ); cdecl;
    self_destuct: procedure( cinfo: Pointer ); cdecl;

    (* Limit on memory allocation for this JPEG object.  (Note that this is
     * merely advisory, not a guaranteed maximum; it only affects the space
     * used for virtual-array buffers.)  May be changed by outer application
     * after creating the JPEG object.
     *)
    max_memory_to_use: Integer;

    //* Maximum allocation request accepted by alloc_large. */
    max_alloc_chunk: Integer;
  end;

  Pjpeg_memory_mgr = ^jpeg_memory_mgr;

//* Progress monitor object */

  jpeg_progress_mgr = record
    progress_monitor: procedure( cinfo: Pointer ); cdecl;

    pass_counter: Integer;	//* work units completed in this pass */
    pass_limit: Integer;	//* total number of work units in this pass */
    completed_passes: Integer;	//* passes completed so far */
    total_passes: Integer;	//* total number of passes expected */
    // extra Delphi info
    instance: PKOLJpeg;         // ptr to current PKOLJpeg object
    last_pass: Integer;
    last_pct: Integer;
    last_time: Integer;
    last_scanline: Integer;
  end;

  Pjpeg_progress_mgr = ^jpeg_progress_mgr;

//* Data source object for decompression */
  JOCTET = BYTE;
  PJOCTET = ^JOCTET;

  jpeg_source_mgr = record
    next_input_byte: PJOCTET;     //* => next byte to read from buffer */
    bytes_in_buffer: DWORD;	//* # of bytes remaining in buffer */

    init_source: procedure( cinfo: Pointer ); cdecl;
    fill_input_buffer: function( cinfo: Pointer ): Boolean; cdecl;
    skip_input_data: procedure( cinfo: Pointer; num_bytes: Integer ); cdecl;
    resync_to_restart: function( cinfo: Pointer; desired: Integer ): Boolean; cdecl;
    term_source: procedure( cinfo: Pointer ); cdecl;
  end;

  Pjpeg_source_mgr = ^jpeg_source_mgr;


//---------------------------------------------//
//* Master record for a decompression instance */
  jpeg_decompress_struct = record
    //jpeg_common_fields;		//* Fields shared with jpeg_compress_struct */
    err: Pjpeg_error_mgr;	        //* Error handler module */\
    mem: Pjpeg_memory_mgr;	//* Memory manager module */\
    progress: Pjpeg_progress_mgr; //* Progress monitor, or NULL if none */\
    client_data: Pointer;		//* Available for use by application */\
    is_decompressor: Boolean;	//* So common code can tell which is which */\
    global_state: Integer;	//* For checking call sequence validity */

    //* Source of compressed data */
    src: Pjpeg_source_mgr;

    //* Basic description of image --- filled in by jpeg_read_header(). */
    //* Application may inspect these values to decide how to process image. */

    image_width: DWORD;	        //* nominal image width (from SOF marker) */
    image_height: DWORD;	        //* nominal image height */
    num_components: Integer;	//* # of color components in JPEG image */
    jpeg_color_space: J_COLOR_SPACE; //* colorspace of JPEG image */
    dummy_align1: array[0..2] of BYTE;

    (* Decompression processing parameters --- these fields must be set before
     * calling jpeg_start_decompress().  Note that jpeg_read_header() initializes
     * them to default values.
     *)

    out_color_space: J_COLOR_SPACE; //* colorspace for output */

    scale_num, scale_denom: DWORD; //* fraction by which to scale image */

    output_gamma: Double;		//* image gamma wanted in output */

    buffered_image: Boolean;	//* TRUE=multiple output passes */
    raw_data_out: Boolean;	//* TRUE=downsampled data wanted */
    dummy_align2: Word;

    dct_method: J_DCT_METHOD;	//* IDCT algorithm selector */
    dummy_align3: array[ 0..2 ] of Byte;
    
    do_fancy_upsampling: Boolean;	//* TRUE=apply fancy upsampling */
    do_block_smoothing: Boolean;	//* TRUE=apply interblock smoothing */

    quantize_colors: Boolean;	//* TRUE=colormapped output wanted */
    dummy_align4: Byte;
    //* the following are ignored if not quantize_colors: */
    dither_mode: J_DITHER_MODE;	//* type of color dithering to use */
    dummy_align5: array[ 0..2 ] of Byte;
    two_pass_quantize: Boolean;	//* TRUE=use two-pass color quantization */
    desired_number_of_colors: Integer;	//* max # colors to use in created colormap */
    //* these are significant only in buffered-image mode: */
    enable_1pass_quant: Boolean;	//* enable future use of 1-pass quantizer */
    enable_external_quant: Boolean;//* enable future use of external colormap */
    enable_2pass_quant: Boolean;	//* enable future use of 2-pass quantizer */

    (* Description of actual output image that will be returned to application.
     * These fields are computed by jpeg_start_decompress().
     * You can also use jpeg_calc_output_dimensions() to determine these values
     * in advance of calling jpeg_start_decompress().
     *)

    output_width: DWORD;	        //* scaled image width */
    output_height: DWORD;	        //* scaled image height */
    out_color_components: Integer;//* # of color components in out_color_space */
    output_components: Integer;	//* # of color components returned */
    (* output_components is 1 (a colormap index) when quantizing colors;
     * otherwise it equals out_color_components.
     *)
    rec_outbuf_height: Integer;	//* min recommended height of scanline buffer */
    (* If the buffer passed to jpeg_read_scanlines() is less than this many rows
     * high, space and time will be wasted due to unnecessary data copying.
     * Usually rec_outbuf_height will be 1 or 2, at most 4.
     *)

    (* When quantizing colors, the output colormap is described by these fields.
     * The application can supply a colormap by setting colormap non-NULL before
     * calling jpeg_start_decompress; otherwise a colormap is created during
     * jpeg_start_decompress or jpeg_start_output.
     * The map has out_color_components rows and actual_number_of_colors columns.
     *)
    actual_number_of_colors: Integer;//* number of entries in use */
    colormap: PJSAMPARRAY;	     //* The color map as a 2-D pixel array */

    (* State variables: these variables indicate the progress of decompression.
     * The application may examine these but must not modify them.
     *)

    (* Row index of next scanline to be read from jpeg_read_scanlines().
     * Application may use this to control its processing loop, e.g.,
     * "while (output_scanline < output_height)".
     *)
    output_scanline: DWORD;	//* 0 .. output_height-1  */

    (* Current input scan number and number of iMCU rows completed in scan.
     * These indicate the progress of the decompressor input side.
     *)
    input_scan_number: Integer;	//* Number of SOS markers seen so far */
    input_iMCU_row: Integer;	//* Number of iMCU rows completed */

    (* The "output scan number" is the notional scan being displayed by the
     * output side.  The decompressor will not allow output scan/row number
     * to get ahead of input scan/row, but it can fall arbitrarily far behind.
     *)
    output_scan_number: Integer;	//* Nominal scan number being displayed */
    output_iMCU_row: DWORD;	//* Number of iMCU rows read */

    (* Current progression status.  coef_bits[c][i] indicates the precision
     * with which component c's DCT coefficient i (in zigzag order) is known.
     * It is -1 when no data has yet been received, otherwise it is the point
     * transform (shift) value for the most recent scan of the coefficient
     * (thus, 0 at completion of the progression).
     * This pointer is NULL when reading a non-progressive file.
     *)
    coef_bits: Pointer;	//* -1 or current Al value for each coef */

    (* Internal JPEG parameters --- the application usually need not look at
     * these fields.  Note that the decompressor output side may not use
     * any parameters that can change between scans.
     *)

    (* Quantization and Huffman tables are carried forward across input
     * datastreams when processing abbreviated JPEG datastreams.
     *)

    quant_tbl_ptrs: array[0..NUM_QUANT_TBLS-1] of PJQUANT_TBL;
    //* ptrs to coefficient quantization tables, or NULL if not defined */

    dc_huff_tbl_ptrs: array[0..NUM_HUFF_TBLS-1] of PJHUFF_TBL;
    ac_huff_tbl_ptrs: array[0..NUM_HUFF_TBLS-1] of PJHUFF_TBL;
    //* ptrs to Huffman coding tables, or NULL if not defined */

    (* These parameters are never carried across datastreams, since they
     * are given in SOF/SOS markers or defined to be reset by SOI.
     *)

    data_precision: Integer;   	//* bits of precision in image data */

    comp_info: Pjpeg_component_info;
    //* comp_info[i] describes component that appears i'th in SOF */

    progressive_mode: Boolean;	//* TRUE if SOFn specifies progressive mode */
    arith_code: Boolean;		//* TRUE=arithmetic coding, FALSE=Huffman */

    arith_dc_L: array[0..NUM_ARITH_TBLS-1] of BYTE; //* L values for DC arith-coding tables */
    arith_dc_U: array[0..NUM_ARITH_TBLS-1] of BYTE; //* U values for DC arith-coding tables */
    arith_ac_K: array[0..NUM_ARITH_TBLS-1] of BYTE; //* Kx values for AC arith-coding tables */

    restart_interval: DWORD;      //* MCUs per restart interval, or 0 for no restart */

    (* These fields record data obtained from optional markers recognized by
     * the JPEG library.
     *)
    saw_JFIF_marker: Boolean;	//* TRUE iff a JFIF APP0 marker was found */
    //* Data copied from JFIF marker; only valid if saw_JFIF_marker is TRUE: */
    JFIF_major_version: BYTE;	//* JFIF version number */
    JFIF_minor_version: BYTE;
    density_unit: BYTE;		//* JFIF code for pixel size units */
    X_density: WORD;		//* Horizontal pixel density */
    Y_density: WORD;		//* Vertical pixel density */
    saw_Adobe_marker: Boolean;	//* TRUE iff an Adobe APP14 marker was found */
    Adobe_transform: BYTE;	//* Color transform code from Adobe marker */

    CCIR601_sampling: Boolean;	//* TRUE=first samples are cosited */

    (* Aside from the specific data retained from APPn markers known to the
     * library, the uninterpreted contents of any or all APPn and COM markers
     * can be saved in a list for examination by the application.
     *)
    marker_list: {jpeg_saved_marker_ptr} Pointer; //* Head of list of saved markers */

    (* Remaining fields are known throughout decompressor, but generally
     * should not be touched by a surrounding application.
     *)

    (*
     * These fields are computed during decompression startup
     *)
    max_h_samp_factor: Integer;	//* largest h_samp_factor */
    max_v_samp_factor: Integer;	//* largest v_samp_factor */

    min_DCT_scaled_size: Integer;	//* smallest DCT_scaled_size of any component */

    total_iMCU_rows: DWORD;	//* # of iMCU rows in image */
    (* The coefficient controller's input and output progress is measured in
     * units of "iMCU" (interleaved MCU) rows.  These are the same as MCU rows
     * in fully interleaved JPEG scans, but are used whether the scan is
     * interleaved or not.  We define an iMCU row as v_samp_factor DCT block
     * rows of each component.  Therefore, the IDCT output contains
     * v_samp_factor*DCT_scaled_size sample rows of a component per iMCU row.
     *)

    sample_range_limit: {PJSAMPLE} Pointer; //* table for fast range-limiting */

    (*
     * These fields are valid during any one scan.
     * They describe the components and MCUs actually appearing in the scan.
     * Note that the decompressor output side must not use these fields.
     *)
    comps_in_scan: Integer;	//* # of JPEG components in this scan */
    cur_comp_info: array[0..MAX_COMPS_IN_SCAN-1] of Pjpeg_component_info;
    //* *cur_comp_info[i] describes component that appears i'th in SOS */

    MCUs_per_row: DWORD;	        //* # of MCUs across the image */
    MCU_rows_in_scan: DWORD;      //* # of MCU rows in the image */

    blocks_in_MCU: Integer;	//* # of DCT blocks per MCU */
    MCU_membership: array[0..D_MAX_BLOCKS_IN_MCU-1] of Integer;
    //* MCU_membership[i] is index in cur_comp_info of component owning */
    //* i'th block in an MCU */

    Ss, Se, Ah, Al: Integer;	//* progressive JPEG parameters for scan */

    (* This field is shared between entropy decoder and marker parser.
     * It is either zero or the code of a JPEG marker that has been
     * read from the data source, but has not yet been processed.
     *)
    unread_marker: Integer;

    (*
     * Links to decompression subobjects (methods, private variables of modules)
     *)
    master: {Pjpeg_decomp_master} Pointer;
    main: {Pjpeg_d_main_controller} Pointer;
    coef: {Pjpeg_d_coef_controller} Pointer;
    post: {Pjpeg_d_coef_controller} Pointer;
    inputctl: {Pjpeg_input_controller} Pointer;
    marker: {Pjpeg_marker_reader} Pointer;
    entropy: {Pjpeg_entropy_decoder} Pointer;
    idct: {Pjpeg_inverse_dct} Pointer;
    upsample: {Pjpeg_upsampler} Pointer;
    cconvert: {Pjpeg_color_deconverter} Pointer;
    cquantize: {Pjpeg_color_quantizer} Pointer;
    dummy_align6: DWORD;
  end;

  Tjpeg_decompress_struct = jpeg_decompress_struct;
  Pjpeg_decompress_struct = ^Tjpeg_decompress_struct;

  { Fields shared with jpeg_decompress_struct }
  jpeg_common_struct = record
    err : Pjpeg_error_mgr;        { Error handler module }
    mem : Pjpeg_memory_mgr;          { Memory manager module }
    progress : Pjpeg_progress_mgr;   { Progress monitor, or NIL if none }
    client_data: Pointer;            { Available for use by application }
    is_decompressor : LongBool;      { so common code can tell which is which }
    global_state : Integer;          { for checking call sequence validity }
  end;

  Pjpeg_common_struct = ^jpeg_common_struct;
  Pjpeg_compress_struct = ^jpeg_compress_struct;

{ Data destination object for compression }
  Pjpeg_destination_mgr = ^jpeg_destination_mgr;
  jpeg_destination_mgr = record
    next_output_byte : PJOCTET;  { => next byte to write in buffer }
    free_in_buffer : Longint;    { # of byte spaces remaining in buffer }

    init_destination : procedure (cinfo : Pjpeg_compress_struct);
    empty_output_buffer : function (cinfo : Pjpeg_compress_struct) : LongBool;
    term_destination : procedure (cinfo : Pjpeg_compress_struct);
  end;

{ Master record for a compression instance }

  jpeg_compress_struct = record
    common: jpeg_common_struct;

    dest : Pjpeg_destination_mgr; { Destination for compressed data }

  { Description of source image --- these fields must be filled in by
    outer application before starting compression.  in_color_space must
    be correct before you can even call jpeg_set_defaults(). }

    image_width : DWORD;         { input image width }
    image_height : DWORD;        { input image height }
    input_components : Integer;       { # of color components in input image }
    in_color_space : J_COLOR_SPACE;   { colorspace of input image }
    dummy_align1: DWORD;
    input_gamma : Double;             { image gamma of input image }

    // Compression parameters
    data_precision : Integer;             { bits of precision in image data }
    num_components : Integer;             { # of color components in JPEG image }
    jpeg_color_space : J_COLOR_SPACE;     { colorspace of JPEG image }
    comp_info : Pointer;
    quant_tbl_ptrs: Array[0..NUM_QUANT_TBLS-1] of Pointer;
    dc_huff_tbl_ptrs : Array[0..NUM_HUFF_TBLS-1] of Pointer;
    ac_huff_tbl_ptrs : Array[0..NUM_HUFF_TBLS-1] of Pointer;
    arith_dc_L : Array[0..NUM_ARITH_TBLS-1] of BYTE; { L values for DC arith-coding tables }
    arith_dc_U : Array[0..NUM_ARITH_TBLS-1] of BYTE; { U values for DC arith-coding tables }
    arith_ac_K : Array[0..NUM_ARITH_TBLS-1] of BYTE; { Kx values for AC arith-coding tables }
    num_scans : Integer;		 { # of entries in scan_info array }
    scan_info : Pointer;     { script for multi-scan file, or NIL }
    raw_data_in : Boolean;        { TRUE=caller supplies downsampled data }
    arith_code : Boolean;         { TRUE=arithmetic coding, FALSE=Huffman }
    optimize_coding : Boolean;    { TRUE=optimize entropy encoding parms }
    CCIR601_sampling : Boolean;   { TRUE=first samples are cosited }
    smoothing_factor : Integer;       { 1..100, or 0 for no input smoothing }
    dct_method : J_DCT_METHOD;    { DCT algorithm selector }
    restart_interval : UINT;      { MCUs per restart, or 0 for no restart }
    restart_in_rows : Integer;        { if > 0, MCU rows per restart interval }

    { Parameters controlling emission of special markers. }
    write_JFIF_header : Boolean;  { should a JFIF marker be written? }
    JFIF_major_version: BYTE;      { What to write for a JFIF version number }
    JFIF_minor_version: BYTE;
    { These three values are not used by the JPEG code, merely copied }
    { into the JFIF APP0 marker.  density_unit can be 0 for unknown, }
    { 1 for dots/inch, or 2 for dots/cm.  Note that the pixel aspect }
    { ratio is defined by X_density/Y_density even when density_unit=0. }
    density_unit : BYTE;         { JFIF code for pixel size units }
    X_density : WORD;           { Horizontal pixel density }
    Y_density : WORD;           { Vertical pixel density }
    write_Adobe_marker : LongBool; { should an Adobe marker be written? }

    { State variable: index of next scanline to be written to
      jpeg_write_scanlines().  Application may use this to control its
      processing loop, e.g., "while (next_scanline < image_height)". }

    next_scanline : DWORD;   { 0 .. image_height-1  }

    { Remaining fields are known throughout compressor, but generally
      should not be touched by a surrounding application. }
    progressive_mode : LongBool;   { TRUE if scan script uses progressive mode }
    max_h_samp_factor : Integer;      { largest h_samp_factor }
    max_v_samp_factor : Integer;      { largest v_samp_factor }
    total_iMCU_rows : DWORD; { # of iMCU rows to be input to coef ctlr }
    comps_in_scan : Integer;          { # of JPEG components in this scan }
    cur_comp_info : Array[0..MAX_COMPS_IN_SCAN-1] of Pointer;
    MCUs_per_row : DWORD;    { # of MCUs across the image }
    MCU_rows_in_scan : DWORD;{ # of MCU rows in the image }
    blocks_in_MCU : Integer;          { # of DCT blocks per MCU }
    MCU_membership : Array[0..C_MAX_BLOCKS_IN_MCU-1] of Integer;
    Ss, Se, Ah, Al : Integer;         { progressive JPEG parameters for scan }

    { Links to compression subobjects (methods and private variables of modules) }
    master : Pointer;
    main : Pointer;
    prep : Pointer;
    coef : Pointer;
    marker : Pointer;
    cconvert : Pointer;
    downsample : Pointer;
    fdct : Pointer;
    entropy : Pointer;
    script_space: Pointer;  { workspace for jpeg_simple_progression }
    script_space_size: Integer;
  end;

  TJPEGContext = record
    err: jpeg_error_mgr;
    progress: jpeg_progress_mgr;
    FinalDCT: J_DCT_METHOD;
    FinalTwoPassQuant: Boolean;
    FinalDitherMode: J_DITHER_MODE;
    case byte of
      0: (common: jpeg_common_struct);
      1: (d: jpeg_decompress_struct);
      2: (c: jpeg_compress_struct);
  end;

procedure ReleaseContext(var jc: TJPEGContext);
begin
  if jc.d.err = nil then Exit;
  j__jpeg_destroy(Integer( @ jc.d ));
  jc.d.err := nil;
end;

function FillInputBuffer( cinfo: Pjpeg_decompress_struct ): Boolean; cdecl;
const EOI: array[ 0..1 ] of Byte = ( $FF, $D9 );
var Strm: PStream;
begin
  Strm := cinfo.client_data;
  cinfo.src.next_input_byte := Pointer( DWORD( Strm.Memory ) + Strm.Position );
  cinfo.src.bytes_in_buffer := Strm.Size - Strm.Position;
  if cinfo.src.bytes_in_buffer = 0 then
  begin
    cinfo.src.next_input_byte := @ EOI[ 0 ];
    cinfo.src.bytes_in_buffer := 2;
  end
    else Strm.Position := Strm.Size;
  Result := TRUE;
end;

procedure SkipInputData( cinfo: Pjpeg_decompress_struct; n: DWORD ); cdecl;
//var Strm: PStream;
begin
  if cinfo.src.bytes_in_buffer >= n then
  begin
    dec( cinfo.src.bytes_in_buffer, n );
    inc( cinfo.src.next_input_byte, n );
  end
    else
  begin
    cinfo.src.bytes_in_buffer := 0;
    {inc( cinfo.src.next_input_byte, n );
    Strm := cinfo.client_data;
    Strm.Position := Strm.Position + n;}
    FillInputBuffer( cinfo );
  end;
end;

procedure TermSource( cinfo: Pjpeg_decompress_struct ); cdecl;
begin
  // nothing
end;

procedure InitSource( cinfo: Pjpeg_decompress_struct ); cdecl;
begin
  // nothing
end;

procedure PrepareSource( cinfo: Pjpeg_decompress_struct );
begin
  cinfo.src.fill_input_buffer := @ FillInputBuffer;
  cinfo.src.skip_input_data := @ SkipInputData;
  cinfo.src.term_source := @ TermSource;
  cinfo.src.init_source := @ InitSource;
end;

procedure InitDecompressor(Obj: PKOLJpeg; var jc: TJPEGContext);
begin
  FillChar(jc, sizeof(jc), #0);
  //jc.err := jpeg_std_error;
  j__jpeg_std_error( Integer( @ jc.err ) );
  jc.d.err := @ jc.err;

  j__jpeg_CreateDecompress( Integer( @ jc.d ), JPEG_LIB_VERSION, sizeof(jc.d) );
  with Obj^ do
  begin
    jc.progress.progress_monitor := FCallback;
    jc.progress.instance := Obj;
    jc.d.progress := @ jc.progress;

    Obj.FImage.FData.Position := 0;
    j__jpeg_stdio_src(Integer( @ jc.d ), Integer( FImage.FData ));
    jc.d.client_data := FImage.FData;
    PrepareSource( @ jc.d );

    j__jpeg_read_header(Integer( @ jc.d ), Integer( TRUE ));

    jc.d.scale_num := 1;
    jc.d.scale_denom := 1 shl Byte(FScale);
    jc.d.do_block_smoothing := FSmoothing;

    if FGrayscale then
    begin
      jc.d.out_color_space := JCS_GRAYSCALE;
      jc.d.quantize_colors := True;
      jc.d.desired_number_of_colors := 236;
    end;

    if FPerformance = jpBestSpeed then
    begin
      jc.d.dct_method := JDCT_IFAST;
      jc.d.two_pass_quantize := False;
      //todo:
      jc.d.do_fancy_upsampling := False;    //!! AV inside jpeglib ??
      jc.d.dither_mode := JDITHER_ORDERED;
    end;

    jc.FinalDCT := jc.d.dct_method;
    jc.FinalTwoPassQuant := jc.d.two_pass_quantize;
    jc.FinalDitherMode := jc.d.dither_mode;
    if FProgressiveDisplay and
       (j__jpeg_has_multiple_scans(Integer( @ jc.d )) <> 0) then
    begin  // save requested settings, reset for fastest on all but last scan
      jc.d.enable_2pass_quant := jc.d.two_pass_quantize;
      jc.d.dct_method := JDCT_IFAST;
      jc.d.two_pass_quantize := False;
      jc.d.dither_mode := JDITHER_ORDERED;
      jc.d.buffered_image := True;
    end;
  end;
end;

procedure InitDestination( cinfo: Pjpeg_compress_struct ); cdecl;
var Strm: PStream;
    Buf: Pointer;
begin
  Strm := cinfo.common.client_data;
  Buf := Pointer( Strm.Tag );
  cinfo.dest.next_output_byte := Buf;
  cinfo.dest.free_in_buffer := 4096;
end;

procedure FlushOutput( cinfo: Pjpeg_compress_struct; len: Integer = 0 );
var Strm: PStream;
    Buf: Pointer;
begin
  Strm := cinfo.common.client_data;
  Buf := Pointer( Strm.Tag );
  if len = 0 then
    len := DWORD( cinfo.dest.next_output_byte ) - DWORD( Buf );
  if len > 0 then
    Strm.Write( Buf^, len );
  cinfo.dest.next_output_byte := Buf;
  cinfo.dest.free_in_buffer := 4096;
end;

procedure EmptyOutputBuffer( cinfo: Pjpeg_compress_struct ); cdecl;
begin
  FlushOutput( cinfo, 4096 );
end;

procedure TermDestination( cinfo: Pjpeg_compress_struct ); cdecl;
var Strm: PStream;
    Buf: Pointer;
begin
  Strm := cinfo.common.client_data;
  Buf := Pointer( Strm.Tag );
  FlushOutput( cinfo );
  FreeMem( Buf );
  Strm.Tag := 0;
end;

procedure PrepareDestination( cinfo: Pjpeg_compress_struct; Strm: PStream );
begin
  cinfo.common.client_data := Strm;
  cinfo.dest.init_destination := @ InitDestination;
  cinfo.dest.empty_output_buffer := @ EmptyOutputBuffer;
  cinfo.dest.term_destination := @ TermDestination;
  Strm.Tag := DWORD( AllocMem( 4096 ) );
end;

const
  JPOOL_PERMANENT = 0;   // lasts until master record is destroyed
  JPOOL_IMAGE	  = 1;	 // lasts until done with image/datastream

{ Decompression startup: read start of JPEG datastream to see what's there
   function jpeg_read_header (cinfo : j_decompress_ptr;
                              require_image : LongBool) : Integer;
  Return value is one of: }
const
  JPEG_SUSPENDED              = 0; { Suspended due to lack of input data }
  JPEG_HEADER_OK              = 1; { Found valid image datastream }
  JPEG_HEADER_TABLES_ONLY     = 2; { Found valid table-specs-only datastream }
{ If you pass require_image = TRUE (normal case), you need not check for
  a TABLES_ONLY return code; an abbreviated file will cause an error exit.
  JPEG_SUSPENDED is only possible if you use a data source module that can
  give a suspension return (the stdio source module doesn't). }


{ function jpeg_consume_input (cinfo : j_decompress_ptr) : Integer;
  Return value is one of: }

  JPEG_REACHED_SOS            = 1; { Reached start of new scan }
  JPEG_REACHED_EOI            = 2; { Reached end of image }
  JPEG_ROW_COMPLETED          = 3; { Completed one iMCU row }
  JPEG_SCAN_COMPLETED         = 4; { Completed last iMCU row of a scan }

function BuildPalette(const cinfo: jpeg_decompress_struct): HPalette;
var
  Pal: TMaxLogPalette;
  I: Integer;
  C: Byte;
begin
  Pal.palVersion := $300;
  Pal.palNumEntries := cinfo.actual_number_of_colors;
  if cinfo.out_color_space = JCS_GRAYSCALE then
    for I := 0 to Pal.palNumEntries-1 do
    begin
      C := cinfo.colormap^[0]^[I];
      Pal.palPalEntry[I].peRed := C;
      Pal.palPalEntry[I].peGreen := C;
      Pal.palPalEntry[I].peBlue := C;
      Pal.palPalEntry[I].peFlags := 0;
    end
  else
    for I := 0 to Pal.palNumEntries-1 do
    begin
      Pal.palPalEntry[I].peRed := cinfo.colormap^[0]^[I];
      Pal.palPalEntry[I].peGreen := cinfo.colormap^[1]^[I];
      Pal.palPalEntry[I].peBlue := cinfo.colormap^[2]^[I];
      // 23 Jun 2005: R <-> B fixed (reporter: Sapersky)
      Pal.palPalEntry[I].peFlags := 0;
    end;
  Result := CreatePalette(PLogPalette(@Pal)^);
end;

procedure BuildColorMap(var cinfo: jpeg_decompress_struct; P: HPalette);
var
  Pal: TMaxLogPalette;
  Count, I: Integer;
begin
  Count := GetPaletteEntries(P, 0, 256, Pal.palPalEntry);
  if Count = 0 then Exit;       // jpeg_destroy will free colormap
  cinfo.colormap := cinfo.mem.alloc_sarray(@ cinfo, JPOOL_IMAGE,
    Count, 3);
  cinfo.actual_number_of_colors := Count;
  for I := 0 to Count-1 do
  begin
    Byte(cinfo.colormap^[2]^[I]) := Pal.palPalEntry[I].peRed;
    Byte(cinfo.colormap^[1]^[I]) := Pal.palPalEntry[I].peGreen;
    Byte(cinfo.colormap^[0]^[I]) := Pal.palPalEntry[I].peBlue;
  end;
end;

procedure SetBitmapDIBPalette( Bmp: PBitmap; Pal: HPalette );
var Entries: array[ 0..255 ] of Integer;
    I: Integer;
begin
  GetPaletteEntries( Pal, 0, 256, Entries[ 0 ] );
  for I := 0 to 255 do
    Bmp.DIBPalEntries[ I ] := Entries[ I ];
end;

procedure DoConvertCMYK2RGB( Bmp: PBitmap; Y_start, Y_end: Integer );
var I, J: Integer;
    C, M, Y, K, R, G, B: Integer;
    P: PDWORD;
begin
  if Bmp.PixelFormat <> pf32bit then Exit;
  for I := Y_start to Y_end do
  begin
    P := Bmp.ScanLine[ I ];
    for J := 0 to Bmp.Width - 1 do
    begin
      C := P^ and $FF;
      M := (P^ shr 8) and $FF;
      Y := (P^ shr 16) and $FF;
      K := P^ shr 24;
      R := Y * K div 255;
      G := M * K div 255;
      B := C * K div 255;
      P^ := R or (G shl 8) or (B shl 16);
      Inc( P );
    end;
  end;
end;

function NewJpeg: PKOLJpeg;
begin
  new( Result, Create );
end;

function DummyProgress( JPEG_obj: PKOLJpeg; const R: TRect ): Boolean;
begin
  Result := FALSE; // not stop
end;

function NormalProgress( JPEG_obj: PKOLJpeg; const R: TRect ): Boolean;
begin
  Result := FALSE; // not stop
  if Assigned( JPEG_obj.FOnProgress ) then
    JPEG_obj.FOnProgress( JPEG_obj, R, Result );
end;

procedure DummyProgressCallback( const cinfo: jpeg_common_struct );
begin
  // * nothing *
end;

procedure ProgressCallback(const cinfo: jpeg_common_struct); cdecl;
var
  Ticks: Integer;
  R: TRect;
  temp: Integer;
begin
  if (cinfo.progress = nil) or (cinfo.progress^.instance = nil) then Exit;
  with cinfo.progress^ do
  begin
    Ticks := GetTickCount;
    if (Ticks - last_time) < cinfo.progress^.instance.fProgressTime then Exit;
    temp := last_time;
    last_time := Ticks;
    if temp = 0 then Exit;
    if cinfo.is_decompressor then
      with Pjpeg_decompress_struct(@cinfo)^ do
      begin
        R := MakeRect(0, last_scanline, output_width, output_scanline);
        if R.Bottom < last_scanline then
          R.Bottom := output_height;
      end
    else
      R := MakeRect(0,0,0,0);
    temp := Integer(Trunc(100.0*(completed_passes + (pass_counter/pass_limit))/total_passes));
    if temp = last_pct then Exit;
    last_pct := temp;
    if cinfo.is_decompressor then
      last_scanline := Pjpeg_decompress_struct(@cinfo)^.output_scanline;
    //instance.Progress(instance, psRunning, temp, (R.Bottom - R.Top) >= 4, R, '');
    if instance.FProgress( instance, R ) then
      instance.FStop := TRUE;
  end;
end;

{ TKOLJpeg }

procedure TKOLJpeg.Changed;
begin
  //todo:
end;

procedure TKOLJpeg.Clear;
begin
  FreeBitmap;
  FImage.Clear;
end;

procedure TKOLJpeg.Compress;
var
  LinesWritten, LinesPerCall: Integer;
  SrcScanLine: Pointer;
  PtrInc: Integer;
  jc: TJPEGContext;
  Src: PBitmap;
  w, h, c: Integer;
begin
  CreateImage;
  FillChar(jc, sizeof(jc), #0);
  j__jpeg_std_error( Integer( @ jc.err ) );
  jc.common.err := @jc.err;

  j__jpeg_CreateCompress(Integer( @ jc.c ), JPEG_LIB_VERSION, sizeof(jc.c));
  try
    try
      jc.progress.progress_monitor := FCallback;
      jc.progress.instance := @Self;
      jc.common.progress := @jc.progress;

      if FImage.FData <> nil then CreateImage;
      FImage.FData := NewMemoryStream;
      FImage.FData.Position := 0;
      j__jpeg_stdio_dest(Integer( @ jc.c ), Integer( FImage.FData ));
      PrepareDestination( @ jc.c, FImage.FData );

      if Assigned( FOnGetScanline ) and (FBitmap <> nil) and (FBitmap.Empty) then Exit
      else
      if Assigned( FBitmap ) and FBitmap.Empty then Exit;
      // if (FBitmap = nil) but Assigned( FOnGetScanline ) then
      // FOnGetScanline will be called for each scanline and also to
      // get image size and attributes
      if Assigned( FBitmap ) then
      begin
        jc.c.image_width := FBitmap.Width;
        FImage.FWidth := FBitmap.Width;
        jc.c.image_height := FBitmap.Height;
        FImage.FHeight := FBitmap.Height;
        jc.c.input_components := 3;           // JPEG requires 24bit RGB input
        jc.c.in_color_space := JCS_RGB;
      end
        else
      begin
        FOnGetScanLine( @ Self, w, h, c, 0, SrcScanLine );
        jc.c.image_width := w;
        FImage.FWidth := w;
        jc.c.image_height := h;
        FImage.FHeight := h;
        jc.c.input_components := Abs( c );           // JPEG requires 24bit RGB input
        CASE c OF
        3: jc.c.in_color_space := JCS_RGB;
        -3: jc.c.in_color_space := JCS_YCbCr;
        1: begin
             jc.c.in_color_space := JCS_GRAYSCALE;
             FImage.FGrayscale := TRUE;
           end;
        4: jc.c.in_color_space := JCS_CMYK;
        -4: jc.c.in_color_space := JCS_YCCK;
        END;
      end;

      if Assigned( FBitmap ) then Src := NewBitmap( 0, 0 )
                             else Src := nil;
      try
        if Src <> nil then
        begin
          Src.Assign(FBitmap);
          Src.PixelFormat := pf24bit;
        end;

        j__jpeg_set_defaults(Integer( @ jc.c ));
        j__jpeg_set_quality(Integer( @ jc.c ), FQuality, Integer( True ));

        if FGrayscale then
        begin
          FImage.FGrayscale := True;
          j__jpeg_set_colorspace(Integer( @ jc.c ), Integer( JCS_GRAYSCALE ));
        end;

        if ProgressiveEncoding then
          j__jpeg_simple_progression(Integer( @ jc.c ));

        if Src <> nil then SrcScanline := Src.ScanLine[0];
        PtrInc := 0;
        if (FImage.FHeight > 1) and Assigned( Src ) then
          PtrInc := Integer(Src.ScanLine[1]) - Integer(SrcScanline);

        // if no dword padding required and source bitmap is top-down
        if (PtrInc > 0) and ((PtrInc and 3) = 0) then
          LinesPerCall := jc.c.image_height  // do whole bitmap in one call
        else
          LinesPerCall := 1;      // otherwise spoonfeed one row at a time

        //--Progress(Self, psStarting, 0, False, Rect(0,0,0,0), '');
        FProgress( @Self, MakeRect( 0, 0, 0, 0 ) );
        try
          j__jpeg_start_compress(Integer( @ jc.c ), Integer( True ));

          while (jc.c.next_scanline < jc.c.image_height) do
          begin
            LinesWritten := j__jpeg_write_scanlines(Integer( @ jc.c ),
                         Integer( @SrcScanline ), LinesPerCall);
            if Src <> nil then
              Inc(Integer(SrcScanline), PtrInc * LinesWritten)
            else
            if jc.c.next_scanline < jc.c.image_height then
            begin
              FOnGetScanLine( @ Self, w, h, c, jc.c.next_scanline, SrcScanLine );
              if w < 0 then break;
            end;
          end;

          j__jpeg_finish_compress(Integer( @ jc.c ));
        finally
          {--
          if ExceptObject = nil then
            PtrInc := 100
          else
            PtrInc := 0;
          --}
          //--Progress(Self, psEnding, PtrInc, False, Rect(0,0,0,0), '');
          //FProgress( @Self, MakeRect( 0, 0, 0, 0 ) );
        end;
      finally
        Src.Free;
      end;
    {$IFDEF JPEGERR}
    except
      //on EAbort do    // OnProgress can raise EAbort to cancel image save
        CreateImage;     // Throw away any partial jpg data
    {$ELSE}
    finally {+}
    {$ENDIF}
    end;
  finally
    ReleaseContext(jc);
  end;
end;

procedure TKOLJpeg.CreateImage;
begin
  FImage.Free;
  new( FImage, Create );
end;

destructor TKOLJpeg.Destroy;
begin
  Clear;
  FImage.Free;
  FBitmap.Free;
  inherited;
end;

procedure TKOLJpeg.DIBNeeded;
var
  LinesPerCall, LinesRead: Integer;
  DestScanLine: Pointer;
  PtrInc: Integer;
  jc: TJPEGContext;
  GeneratePalette: Boolean;
  PaletteModified : Boolean;

  TmpPal: HPalette;
  OK: Boolean;
  w, h: Integer;
  wbmp, hbmp, cbmp, line: Integer;
begin
  if FBitmap <> nil then Exit;
  if FImage = nil then Exit;

  if (Width = 0) or (Height = 0) then
    Exit;
  GeneratePalette := True;

  Jpeg_Error := FALSE;
  FStop := FALSE;
  InitDecompressor(@Self, jc);
  FCorrupted := Jpeg_Error;

  FPixelFormat := jf8Bit;

  try
    try
      // Set the bitmap pixel format

      //--Progress(Self, psStarting, 0, False, Rect(0,0,0,0), '');
      if FProgress( @Self, MakeRect( 0, 0, 0, 0 ) ) then Exit;
      PaletteModified := False;
      OK := FALSE;
      try
        j__jpeg_start_decompress( Integer( @ jc.d ) );
        //{$IFNDEF JPEGERR}
        //if Jpeg_Error then Exit;
        //{$ENDIF}

        // Set bitmap width and height
        w := jc.d.output_width;
        h := jc.d.output_height;

        wbmp := w;
        hbmp := h;
        if Assigned( OnGetScanline ) then
          hbmp := 1;

        cbmp := 1;
        if FGrayscale then
          FBitmap := NewDIBBitmap( wbmp, hbmp, pf8bit )
        else
        begin
          cbmp := 3;
          FPixelFormat := jf24Bit;
          if jc.d.out_color_space in [JCS_CMYK,JCS_YCCK] then
          begin
            FBitmap := NewDIBBitmap( wbmp, hbmp, pf32bit );
            //jc.d.out_color_space := JCS_RGB;
            cbmp := 4;
          end
          else
            FBitmap := NewDIBBitmap( wbmp, hbmp, pf24bit );
        end;

        if (FTempPal <> 0) then
        begin
          if (FBitmap.PixelFormat = pf8Bit) then //??? At this place FBitmap = nil !!!
          begin                        // Generate DIB using assigned palette
            BuildColorMap(jc.d, FTempPal);
            SetBitmapDIBPalette( FBitmap, FTempPal );

            GeneratePalette := False;
          end
          else
          begin
            DeleteObject(FTempPal);
            FTempPal := 0;
          end;
        end;

        with FBitmap^ do
        begin
          DestScanline := ScanLine[0];
          PtrInc := 0;
          if (jc.d.output_height > 1) and (hbmp > 1) then
            PtrInc := Integer(ScanLine[1]) - Integer(DestScanline);
          if (PtrInc > 0) and ((PtrInc and 3) = 0) and (hbmp > 1) then
             // if no dword padding is required and output bitmap is top-down
            LinesPerCall := jc.d.rec_outbuf_height // read multiple rows per call
          else
            LinesPerCall := 1;            // otherwise read one row at a time
        end;

        FCMYK := FALSE;
        if jc.d.out_color_space in [JCS_CMYK,JCS_YCCK] then
          FCMYK := TRUE;

        if jc.d.buffered_image then
        begin  // decode progressive scans at low quality, high speed
          while j__jpeg_consume_input(Integer( @ jc.d ) ) <> JPEG_REACHED_EOI do
          begin
            j__jpeg_start_output(Integer( @ jc.d ), jc.d.input_scan_number);
            if Jpeg_Error then Exit;
            if FStop then Exit;
            // extract color palette
            if (jc.d.progress^.completed_passes = 0) and (jc.d.colormap <> nil)
              and (FBitmap.PixelFormat = pf8bit) and GeneratePalette then
            begin
              //
              //FBitmap.Palette := BuildPalette(jc.d);
              ///////////////////////////////////////////
              TmpPal := BuildPalette(jc.d);            //
              SetBitmapDIBPalette( FBitmap, TmpPal );  //
              DeleteObject( TmpPal );                  //
              ///////////////////////////////////////////
              PaletteModified := True;
            end;
            DestScanLine := FBitmap.ScanLine[0];
            while (jc.d.output_scanline < jc.d.output_height) do
            begin
              line := jc.d.output_scanline;
              LinesRead := j__jpeg_read_scanlines(Integer( @ jc.d ),
                        Integer( @ DestScanline ), LinesPerCall);
              if Assigned( OnGetScanline ) then
              begin
                h := -Abs( h );
                OnGetScanline( @ Self, w, h, cbmp, line, DestScanline );
                h := Abs( h );
                if h < 0 then Exit;
              end;
              if hbmp > 1 then
                Inc(Integer(DestScanline), PtrInc * LinesRead);
            end;
            j__jpeg_finish_output(Integer( @ jc.d ));
          end;
          // reset options for final pass at requested quality
          jc.d.dct_method := jc.FinalDCT;
          jc.d.dither_mode := jc.FinalDitherMode;
          if jc.FinalTwoPassQuant then
          begin
            jc.d.two_pass_quantize := True;
            jc.d.colormap := nil;
          end;
          j__jpeg_start_output(Integer( @jc.d ), jc.d.input_scan_number);
          DestScanLine := FBitmap.ScanLine[0];
        end;

        // build final color palette
        if (not jc.d.buffered_image or jc.FinalTwoPassQuant) and
          (jc.d.colormap <> nil) and GeneratePalette then
        begin
          //
          //FBitmap.Palette := BuildPalette(jc.d);
          ///////////////////////////////////////////
          TmpPal := BuildPalette(jc.d);            //
          SetBitmapDIBPalette( FBitmap, TmpPal );  //
          DeleteObject( TmpPal );                  //
          ///////////////////////////////////////////
          PaletteModified := True;
          DestScanLine := FBitmap.ScanLine[0];
        end;
        // final image pass for progressive, first and only pass for baseline
        while (jc.d.output_scanline < jc.d.output_height) do
        begin
          if FStop then Exit;
          line := jc.d.output_scanline;
          LinesRead := j__jpeg_read_scanlines(Integer( @ jc.d ),
                   Integer( @DestScanline ), LinesPerCall);
          if Assigned( OnGetScanline ) then
          begin
            OnGetScanline( @ Self, w, h, cbmp, line, DestScanline );
            if w < 0 then Exit;
          end;
          if hbmp > 1 then
            Inc(Integer(DestScanline), PtrInc * LinesRead);
        end;

        if jc.d.buffered_image then j__jpeg_finish_output(Integer( @ jc.d ));
        j__jpeg_finish_decompress(Integer( @ jc.d ));
        OK := TRUE;
      finally
        {--
        if ExceptObject = nil then
          PtrInc := 100
        else
          PtrInc := 0;
        --}
        //--Progress(Self, psEnding, PtrInc, PaletteModified, Rect(0,0,0,0), '');
        if OK then
          FProgress( @Self, MakeRect( 0, 0, Width, Height ) );
        // Make sure new palette gets realized, in case OnProgress event didn't.
        if PaletteModified then
          Changed;
        Jpeg_Error := Jpeg_Error or not OK;
        FCMYK := FALSE;
        if jc.d.out_color_space in [JCS_CMYK,JCS_YCCK] then
        begin
          FCMYK := TRUE;
          if Assigned( FConvertCMYK2RGBProc ) then
          begin
            FConvertCMYK2RGBProc( FBitmap, 0, FBitmap.Height-1 );
            FCMYK := FALSE;
          end;
        end;
      end;
    except
      //--on EAbort do ;   // OnProgress can raise EAbort to cancel image load
      {$IFDEF JPEGERR}
      FCorrupted := TRUE;
      {$ENDIF}
    end;
  finally
    ReleaseContext(jc);
    {$IFNDEF JPEGERR}
    FCorrupted := Jpeg_Error;
    {$ENDIF}
  end;

end;

procedure TKOLJpeg.Draw(DC: HDC; x, y: Integer);
begin
  DIBNeeded;
  if Empty then Exit;
  //FBitmap.Draw( DC, x, y );
  BitBlt( DC, x, y, FBitmap.Width, FBitmap.Height,
          FBitmap.Canvas.Handle, 0, 0, SRCCOPY );
end;

function TKOLJpeg.Empty: Boolean;
begin
  Result := (FImage = nil) or (Width <= 0) or (Height <= 0);
end;

procedure TKOLJpeg.FreeBitmap;
begin
  Free_And_Nil( FBitmap );
end;

function TKOLJpeg.GetBitmap: PBitmap;
begin
  if FBitmap = nil then
    DIBNeeded;
  Result := FBitmap;
end;

function TKOLJpeg.HasBitmap: Boolean;
begin
  Result := FBitmap <> nil;
end;

procedure TKOLJpeg.Init;
begin
  CreateImage;
  FQuality := JPEGDefaults.CompressionQuality;
  FGrayscale := JPEGDefaults.Grayscale;
  FPerformance := JPEGDefaults.Performance;
  FPixelFormat := JPEGDefaults.PixelFormat;
  FProgressiveDisplay := JPEGDefaults.ProgressiveDisplay;
  FProgressiveEncoding := JPEGDefaults.ProgressiveEncoding;
  FScale := JPEGDefaults.Scale;
  FSmoothing := JPEGDefaults.Smoothing;
  FProgress := DummyProgress;
  FCallback := @ DummyProgressCallback;
  fProgressTime := 100;
end;

procedure TKOLJpeg.JPEGNeeded;
begin
  if (FImage = nil) or (FImage.FData = nil) then
    Compress;
end;

function TKOLJpeg.LoadFromFile(const filename: String): Boolean;
var Strm: PStream;
begin
  Strm := NewReadFileStream( filename );
  TRY
    Result := LoadFromStream( Strm );
  FINALLY
    Strm.Free;
  END;
end;

function TKOLJpeg.LoadFromStream(Strm: PStream): Boolean;
var jerr: jpeg_error_mgr;
    cinfo: jpeg_decompress_struct;
begin
  Result := FALSE;
  FCorrupted := FALSE;
  CreateImage;
  FreeBitmap;
  FData := NewMemoryStream;
  FData.Size := Strm.Size;
  Strm.Read( FData.Memory^, Strm.Size );
  FData.Position := 0;
  FImage.FData := FData;
  FWidth := 0;
  FHeight := 0;
  if Strm.Size > 0 then
  begin
    //jerr := _jpeg_std_error;
    FillChar( cinfo, Sizeof( cinfo ), #0 );
    j__jpeg_std_error( Integer( @ jerr ) );
    cinfo.err := @ jerr;
    j__jpeg_CreateDecompress( Integer( @ cinfo ), JPEG_LIB_VERSION,
                              sizeof( jpeg_decompress_struct ) );
    TRY
      j__jpeg_stdio_src( Integer( @ cinfo ), Integer( FData ) );
      cinfo.client_data := FData;
      PrepareSource( @ cinfo );
      j__jpeg_read_header( Integer( @ cinfo ), {require_imaged=}Integer( TRUE ) );
      FWidth := cinfo.image_width;
      FHeight := cinfo.image_height;
      FGrayScale := cinfo.jpeg_color_space = JCS_GRAYSCALE;
      FProgressiveEncoding := j__jpeg_has_multiple_scans( Integer( @ cinfo ) ) <> 0;
    FINALLY
      j__jpeg_destroy_decompress( Integer( @ cinfo ) );
    END;
  end;
end;

function TKOLJpeg.SaveToFile(const filename: String): Boolean;
var Strm: PStream;
begin
  Strm := NewWriteFileStream( filename );
  TRY
    Result := SaveToStream( Strm );
  FINALLY
    Strm.Free;
  END;
end;

function TKOLJpeg.SaveToStream(Strm: PStream): Boolean;
begin
  JPEGNeeded;
  if FImage <> nil then
    with FImage.FData^ do
    begin
      Result := (Size > 0) and (Strm.Write(Memory^, Size) = Size);
    end
  else Result := Assigned( OnGetScanline );
end;

procedure TKOLJpeg.SetBitmap(const Value: PBitmap);
begin
  Free_And_Nil( FImage );
  if Value = FBitmap then Exit;
  FreeBitmap;
  if FBitmap = nil then
    FBitmap := NewBitmap( 0, 0 );
  FBitmap.Assign( Value );
end;

procedure TKOLJpeg.SetConvertCMYK2RGB(const Value: Boolean);
begin
  FConvertCMYK2RGB := Value;
  FConvertCMYK2RGBProc := nil;
  if Value then
    FConvertCMYK2RGBProc := DoConvertCMYK2RGB;
end;

procedure TKOLJpeg.SetOnProgress(const Value: TJPEGProgress);
begin
  FOnProgress := Value;
  FProgress := NormalProgress;
  FCallback := @ ProgressCallback;
end;

procedure TKOLJpeg.SetScale(const Value: TJPEGScale);
begin
  if FScale = Value then Exit;
  FScale := Value;
  if (FBitmap <> nil) and (FImage <> nil) then
    FreeBitmap;
end;

procedure TKOLJpeg.StretchDraw(DC: HDC; const R: TRect);
begin
  DIBNeeded;
  if Empty then Exit;
  FBitmap.StretchDraw( DC, R );
end;

{ TJPEGData }

procedure TJPEGData.Clear;
begin
  if FData <> nil then
     FData.Size := 0;
  FWidth := 0;
  FHeight := 0;
end;

destructor TJPEGData.Destroy;
begin
  FData.Free;
  inherited;
end;

end.
